import math
import numpy as np
import numbers
import matplotlib.pyplot as plt

import SInterval as SInt


class Parameters:
    interp_points = 100     # number of interpolation points 
    sim_interp_points = 100 # number of interpolation points in similarity measure calculations
    ZER_points = 100        # number of points in Zadeh extension rule operations
    tolerance = 0.000001    # minimum support length
    
    

class FNumber:
    '''
    Fuzzy Number Class
    '''

    def __init__(self, fn = 2, points = [0, 1], name = ''):
        self.name = name
        self.fntype = fn
        
        if  fn == 1: # singleton
            self.singleton_position = points[0];
            self.mu = [0.0, 1.0]
            self.left = [points[0], points[0]]
            self.right = [points[0], points[0]]
        
        elif fn == 2: # rectangle
            self.singleton_position = 0.5 * (points[0] + points[1]);
            self.mu = [0.0, 1.0]
            self.left = [points[0], points[0]]
            self.right = [points[1], points[1]]
        
        elif fn == 3: # triangle    
            self.singleton_position = points[1];
            self.mu = [0.0, 1.0]
            self.left = [points[0], points[1]]
            self.right = [points[2], points[1]]
        
        elif fn == 4: # trapezoid    
            self.singleton_position = 0.5 * (points[1] + points[2]);
            self.mu = [0.0, 1.0]
            self.left = [points[0], points[1]]
            self.right = [points[3], points[2]]
        
        elif fn == 5: # polynomial
            self.fntype = 0
            self.singleton_position = 0.5 * (points[1] + points[2]);
            self.mu = np.zeros(Parameters.interp_points+1)
            x = np.linspace(0.0, 1.0, Parameters.interp_points+1)
            for i in range(Parameters.interp_points+1):
                self.mu[i] = 1.0 - (x[i] - 1.0)**2
            
            self.left = np.ones_like(self.mu)
            self.right = np.ones_like(self.mu)
            if  points[0]==points[1]:
                self.left = self.left * points[0]
            else:    
                self.left = np.linspace(points[0],points[1],Parameters.interp_points+1)
                
            if  points[2]==points[3]:
                self.right = self.right * points[3]
            else:
                self.right = np.linspace(points[2],points[3],Parameters.interp_points+1)
                self.right = self.right[::-1]
                
        elif fn == 6: # harmonic
            self.fntype = 0
            self.singleton_position = 0.5 * (points[1] + points[2]);
            self.mu = np.zeros(Parameters.interp_points+1)
            x = np.linspace(0.0, 1.0, Parameters.interp_points+1)
            for i in range(Parameters.interp_points+1):
                self.mu[i] = 0.5*(1.0 + math.cos(math.pi*(x[i]-1.0)))
                
            self.left = np.ones_like(self.mu)
            self.right = np.ones_like(self.mu)
            if  points[0]==points[1]:
                self.left = self.left * points[0]
            else:    
                self.left = np.linspace(points[0],points[1],Parameters.interp_points+1)
                
            if  points[2]==points[3]:
                self.right = self.right * points[3]
            else:
                self.right = np.linspace(points[2],points[3],Parameters.interp_points+1)
                self.right = self.right[::-1]
                
        elif fn == 7: # pseudo Gaussian
            self.fntype = 0
            self.singleton_position = 0.5 * (points[1] + points[2]);
            self.mu = np.zeros(Parameters.interp_points+1)
            x = np.linspace(0.0, 1.0, Parameters.interp_points+1)
            for i in range(1,Parameters.interp_points):
                self.mu[i] = math.exp(((2.0-x[i])*x[i] - 1.0)/((2.0-x[i])*x[i]))
                self.mu[Parameters.interp_points] = 1.0
                
            self.left = np.ones_like(self.mu)
            self.right = np.ones_like(self.mu)
            if  points[0]==points[1]:
                self.left = self.left * points[0]
            else:    
                self.left = np.linspace(points[0],points[1],Parameters.interp_points+1)

            if  points[2]==points[3]:
                self.right = self.right * points[3]
            else:
                self.right = np.linspace(points[2],points[3],Parameters.interp_points+1)
                self.right = self.right[::-1]        
            
        else:
            raise NotImplemented()
        
        
    def __repr__(self):
        return self.__str__()
        
    def __str__(self):
        if  self.fntype == 1: # singleton
            return "Singleton fuzzy number '{0.name}' = [{0.left[0]}]".format(self) 
        elif self.fntype == 2: # rectangle
            return "Rectangular fuzzy number '{0.name}' = [{0.left[0]}, {0.right[0]}]".format(self)
        elif self.fntype == 3: # triangle    
            return "Triangular fuzzy number '{0.name}' = [{0.left[0]}, {0.left[1]}, {0.right[0]}]".format(self)
        elif self.fntype == 4: # trapezoid    
            return "Trapezoidal fuzzy number '{0.name}' = [{0.left[0]}, {0.left[1]}, {0.right[1]}, {0.right[0]}]".format(self)
        else:
            return "Fuzzy number '{0.name}'".format(self)
                


    def muValue(self, x):
        mu1 = 0.0
    
        if  self.fntype == 1: # singleton
            if x == self.left[0]: return(1.0)
            else: return(0.0)
        else:
            if  x<self.left[0] or x>self.right[0]: return(0.0)
            elif self.left[-1] <= x <= self.right[-1]: return(1.0)
            else:
                if  self.left[0] <= x < self.left[-1]:
                    mu1 = np.interp(x, self.left, self.mu)
                elif self.right[-1] < x <= self.right[0]:
                    mu1 = np.interp(x, self.right[::-1], self.mu[::-1])
                    
        return(mu1)        
            
            
    def alphaValue(self, mu):
        if  mu<0 or mu>1:
            raise ValueError("mu must be in range [0,1]") 
    
        if  self.fntype == 1: # singleton
            return SInt.SInterval(self.left[0],self.right[0])
     
        elif self.fntype == 2: # rectangle
            return SInt.SInterval(self.left[0],self.right[0])
        
        elif self.fntype == 3: # triangle
            if  mu == 1.0: 
                return SInt.SInterval(self.left[1],self.right[1])
            elif mu == 0.0:
                return SInt.SInterval(self.left[0],self.right[0])
            else:
                vl = np.interp(mu, self.mu, self.left)
                vr = np.interp(mu, self.mu, self.right)
                return SInt.SInterval(vl,vr)
             
        elif self.fntype == 4: # trapezoid
            if  mu == 1.0: 
                return SInt.SInterval(self.left[1],self.right[1])
            elif mu == 0.0:
                return SInt.SInterval(self.left[0],self.right[0])
            else:
                vl = np.interp(mu, self.mu, self.left)
                vr = np.interp(mu, self.mu, self.right)
                return SInt.SInterval(vl,vr)
        else:    
            vl = np.interp(mu, self.mu, self.left)
            vr = np.interp(mu, self.mu, self.right)
            return SInt.SInterval(vl,vr)

              
    def plot(self):
        if  self.fntype == 1: # singleton
            plt.plot([self.left[0],self.left[0]],[0,1])
             
        elif self.fntype == 2: # rectangle
            plt.plot([self.left[0],self.left[0],self.right[0],self.right[0]],[0,1,1,0])
            
        else:
            plt.plot(np.hstack([self.left, self.right[::-1]]), np.hstack([self.mu, self.mu[::-1]]))   
            
        return plt.gca()     
            
    
    def support(self):
        return self.alphaValue(0.0)        
    
    
    def core(self):
        return self.alphaValue(1.0)
    
    
    
    def __neg__(self):
        if  len(self.name) == 0: new_name = ''
        else: new_name = '- ' + self.name
        
        if  self.fntype == 1: # singleton
            return Singleton(-self.left[0], new_name)
        elif self.fntype == 2: # rectangle
            return Rectangle([-self.right[0], -self.left[0]], new_name)
        elif self.fntype == 3: # triangle
            return Triangle([-self.right[0], -self.left[1], -self.left[0]], new_name)
        elif self.fntype == 4: # trapezoid
            return Trapezoid([-self.right[0], -self.right[1], -self.left[1], -self.left[0]], new_name)
        else:
            f = FNumber()
            f.fntype = 0
            f.name = new_name
            f.singleton_position = -self.core().mid()
            f.mu = self.mu
            f.left = self.right*(-1.0)
            f.right = self.left*(-1.0)
            return(f)
        
        
        
    def __add__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            d = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            d  = other
        else:
            raise NotImplemented()    
        
        f = FNumber()
        f.name = '(' + self.name + ') + (' + d.name + ')' 
        f.singleton_position = self.singleton_position + d.singleton_position
        
        if  len(self.mu) == 2 and len(d.mu) == 2:
            f.mu = self.mu
            f.left[0] = self.left[0] + d.left[0]
            f.left[1] = self.left[1] + d.left[1]
            f.right[0] = self.right[0] + d.right[0]
            f.right[1] = self.right[1] + d.right[1]
            if  self.fntype == 1 and d.fntype == 1:
                f.fntype = 1
            elif (self.fntype == 1 and d.fntype == 2) or (self.fntype == 2 and d.fntype == 1) or (self.fntype == 2 and d.fntype == 2):
                f.fntype = 2
            elif (self.fntype == 1 and d.fntype == 3) or (self.fntype == 3 and d.fntype == 1) or (self.fntype == 3 and d.fntype == 3):
                f.fntype = 3
            else:
                f.fntype = 4        

            return(f)
        
        else:
            f.fntype = 0
            f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
            f.left = np.zeros_like(f.mu)
            f.right = np.zeros_like(f.mu)
            for i in range(Parameters.interp_points+1):
                a = self.alphaValue(f.mu[i])
                b = d.alphaValue(f.mu[i])
                c = a + b
                f.left[i] = c.min_val
                f.right[i] = c.max_val
            return(f)
            
        
        
    def __radd__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        else:
            raise NotImplemented()
        
        return self.__add__(d)   



    def __sub__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            d = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            d  = other
        else:
            raise NotImplemented()    
        
        return self + (-d)
        
        

    def __rsub__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        else:
            raise NotImplemented()
        
        return d + (-self)
    
        
        
    def __mul__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            d = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            d  = other
        else:
            raise NotImplemented()    
        
        f = FNumber()
        f.name = '(' + self.name + ') * (' + d.name + ')' 
        f.singleton_position = self.singleton_position * d.singleton_position
        
        # multiply by singleton 
        if  self.fntype == 1 and d.fntype != 1:
            self, d = d, self # zamiana miejscami !
        if  d.fntype == 1:
            f.fntype = self.fntype
            f.mu = self.mu
            if  d.singleton_position >= 0.0:
                f.left = np.array(self.left) * d.singleton_position
                f.right = np.array(self.right) * d.singleton_position
            else:
                f.left = np.array(self.right) * d.singleton_position
                f.right = np.array(self.left) * d.singleton_position    
            return(f)    
            
        f.fntype = 0
        f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
        f.left = np.zeros_like(f.mu)
        f.right = np.zeros_like(f.mu)
        for i in range(Parameters.interp_points+1):
            a = self.alphaValue(f.mu[i])
            b = d.alphaValue(f.mu[i])
            c = a * b
            f.left[i] = c.min_val
            f.right[i] = c.max_val
        return(f)   
        
        
        
    def __rmul__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(other, '{0}'.format(other))
        else:
            raise NotImplemented()
        
        return self.__mul__(d)    
        
        
        
    def __truediv__(self, other):
        if  isinstance(other,numbers.Number):
            if  other == 0.0:
                raise ValueError('Division by 0!')
            d = Singleton(float(other), '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            if other.min_val <= 0.0 <= other.max_val:
                raise ValueError("0 can't belong to interval in denominator!")
            d = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            if  other.left[0] <= 0.0 and other.right[0] >= 0.0:
                raise ValueError("0 can't belong to fuzzy number support in denominator!")
            d  = other
        else:
            raise NotImplemented()    
        
        f = FNumber()
        f.name = '(' + self.name + ') / (' + d.name + ')' 
        f.singleton_position = self.singleton_position / d.singleton_position
        
        # divide by singleton 
        if  d.fntype == 1:
            f.fntype = self.fntype
            f.mu = self.mu
            if  d.singleton_position > 0.0:
                f.left = np.array(self.left) / d.singleton_position
                f.right = np.array(self.right) / d.singleton_position
            else:
                f.left = np.array(self.right) / d.singleton_position
                f.right = np.array(self.left) / d.singleton_position    
            return f
            
        f.fntype = 0
        f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
        f.left = np.zeros_like(f.mu)
        f.right = np.zeros_like(f.mu)
        for i in range(Parameters.interp_points+1):
            a = self.alphaValue(f.mu[i])
            b = d.alphaValue(f.mu[i])
            c = a / b
            f.left[i] = c.min_val
            f.right[i] = c.max_val
        return f
        
        
        
    def __rtruediv__(self, other):
        if  isinstance(other,numbers.Number):
            d = Singleton(float(other), '{0}'.format(other))
        else:
            raise NotImplemented()
        
        return d/self
    
    
    
    def __abs__(self):
        s = self.support()
        if  s.min_val >=0:
            return self
        elif s.max_val <=0:
            return -self
        else:
            if  len(self.name) == 0: new_name = ''
            else: new_name = 'abs(' + self.name + ')'
            
            f = FNumber()
            f.name = new_name 
            f.fntype = 0
            f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
            f.left = np.zeros_like(f.mu)
            f.right = np.zeros_like(f.mu)
            for i in range(Parameters.interp_points+1):
                a = abs(self.alphaValue(f.mu[i]))
                f.left[i] = a.min_val
                f.right[i] = a.max_val
            c = f.core()
            f.singleton_position = c.mid()    
            return f
        
        
        
    def __pow__(self, other):
        if  not isinstance(other,numbers.Number):
            raise ValueError("Exponent must be a number!")
        
        s = self.support()
        if  other < 0 and s.isSubset(0.0):
            raise ValueError("Fuzzy number cannot be raised to a negative power because its support contains 0.0")
        
        if  len(self.name) == 0: new_name = ''
        else: new_name = '(' + self.name + '**' + '{0}'.format(other) + ')'
        
        f = FNumber()
        f.name = new_name 
        f.fntype = 0
        f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
        f.left = np.zeros_like(f.mu)
        f.right = np.zeros_like(f.mu)
        for i in range(Parameters.interp_points+1):
            a = self.alphaValue(f.mu[i]) ** other
            f.left[i] = a.min_val
            f.right[i] = a.max_val
        c = f.core()
        f.singleton_position = c.mid()    
        return f
    
        
        
    def center_of_gravity(self):
        if  self.fntype == 1 or self.fntype == 2:
            return self.singleton_position

        s = self.support()
        if  s.width() <= Parameters.tolerance:
            return self.singleton_position

        if  self.fntype == 3:  # triangle
            a = self.left[0]
            b = self.left[1]
            c = self.right[0]
            return ((a + (2.0/3.0)*(b-a))*(b-a) + (b + (c-b)/3.0)*(c-b))/(c-a)
        
        x = np.linspace(s.min_val, s.max_val, Parameters.interp_points+1)
        dx = x[1] - x[0]
        sumal = 0.0
        sumam = 0.0
        
        for i in range(Parameters.interp_points):
            p = 0.5*(self.muValue(x[i])+self.muValue(x[i+1]))*dx
            sumal += p*0.5*(x[i]+x[i+1])
            sumam += p 
        if  sumam == 0.0:
            w = 0.0
        else:
            w = sumal/sumam
        
        return w



    def mid_point(self):
        if  self.fntype == 1 or self.fntype == 2: # sigleton or rectangular
            return self.singleton_position
        if  self.fntype == 3:  # triangle
            a = self.left[0]
            b = self.left[1]
            c = self.right[0]
            return 0.5 * (0.5*(a+c) + b)
        if  self.fntype == 4:  # trapezoid
            a = self.left[0]
            b = self.left[1]
            c = self.right[1]
            d = self.right[0]
            return 0.5 * ((a+d) + 0.5*(b+c-a-d))

        d = 0.0
        x = np.linspace(0.0, 1.0, Parameters.sim_interp_points + 1)
        dx = x[1] - x[0]
        for i in range(Parameters.sim_interp_points):
            a_fn = self.alphaValue(x[i])
            a_fn_1 = self.alphaValue(x[i + 1])
            d += 0.5 *((a_fn_1.min_val + a_fn_1.max_val) + (a_fn.min_val + a_fn.max_val)) * dx

        return 0.5*d


    
    def ZER_add(self, other):
        if  isinstance(other,numbers.Number):
            b = Singleton(other, '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            b = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            b  = other
        else:
            raise NotImplemented()
        
        a_s = self.support()
        b_s = b.support()
        a_c = self.core()
        b_c = b.core()
        
        l_min = a_s.min_val + b_s.min_val
        l_max = a_c.min_val + b_c.min_val
        r_min = a_c.max_val + b_c.max_val
        r_max = a_s.max_val + b_s.max_val
        
        f = FNumber()
        f.name = '(' + self.name + ') +Z (' + b.name + ')' 
        f.singleton_position = self.singleton_position + b.singleton_position
        f.fntype = 0
        f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
        f.left = np.zeros_like(f.mu)
        f.right = np.zeros_like(f.mu)
                
        c = np.linspace(l_min, l_max, Parameters.ZER_points+1)
        a = np.linspace(a_s.min_val, a_s.max_val, 10*Parameters.ZER_points+1)
        ml = np.zeros(Parameters.ZER_points+1)     
        l =  np.zeros_like(ml)  
        for i in range(Parameters.ZER_points+1):
            l[i] = c[i]
            for j in range(10*Parameters.ZER_points+1):
                bb = c[i] - a[j]
                if  bb < b_s.min_val or bb > b_s.max_val:
                    continue
                else:
                    mm = min(self.muValue(a[j]), b.muValue(bb)) # t-norm operator 
                    ml[i] = max(ml[i], mm)                      # s-norm operator 
        
        c = np.linspace(r_min, r_max, Parameters.ZER_points+1)
        mr = np.zeros(Parameters.ZER_points+1)     
        r =  np.zeros_like(mr)
        for i in range(Parameters.ZER_points+1):
            r[i] = c[i]
            for j in range(10*Parameters.ZER_points+1):
                bb = c[i] - a[j]
                if  bb < b_s.min_val or bb > b_s.max_val:
                    continue
                else:
                    mm = min(self.muValue(a[j]), b.muValue(bb)) # t-norm operator 
                    mr[i] = max(mr[i], mm)                      # s-norm operator 
                                
        mr = mr[::-1]
        r = r[::-1]
        for i in range(Parameters.interp_points+1):
            f.left[i] = np.interp(f.mu[i],ml,l)
            f.right[i] = np.interp(f.mu[i],mr,r)
                       
        return(f)        
            
            
        
    def ZER_mul(self, other):
        if  isinstance(other,numbers.Number):
            b = Singleton(other, '{0}'.format(other))
        elif isinstance(other, SInt.SInterval):
            b = Rectangle([other.min_val, other.max_val], "[{0.min_val},{0.max_val}]".format(other))
        elif isinstance(other, FNumber):
            b  = other
        else:
            raise NotImplemented()
        
        a_s = self.support()
        b_s = b.support()
        a_c = self.core()
        b_c = b.core()
        
        l_min = min(a_s.min_val*b_s.min_val, a_s.min_val*b_s.max_val,a_s.max_val*b_s.min_val,a_s.max_val*b_s.max_val)
        l_max = min(a_c.min_val*b_c.min_val, a_c.min_val*b_c.max_val,a_c.max_val*b_c.min_val,a_c.max_val*b_c.max_val)
        r_min = max(a_c.min_val*b_c.min_val, a_c.min_val*b_c.max_val,a_c.max_val*b_c.min_val,a_c.max_val*b_c.max_val)
        r_max = max(a_s.min_val*b_s.min_val, a_s.min_val*b_s.max_val,a_s.max_val*b_s.min_val,a_s.max_val*b_s.max_val) 
        
        f = FNumber()
        f.name = '(' + self.name + ') *Z (' + b.name + ')' 
        f.singleton_position = self.singleton_position * b.singleton_position
        f.fntype = 0
        f.mu = np.linspace(0.0, 1.0, Parameters.interp_points+1)
        f.left = np.zeros_like(f.mu)
        f.right = np.zeros_like(f.mu)
                
        c = np.linspace(l_min, l_max, Parameters.ZER_points+1)
        a = np.linspace(a_s.min_val, a_s.max_val, 10*Parameters.ZER_points+1)
        ml = np.zeros(Parameters.ZER_points+1)     
        l =  np.zeros_like(ml)  
        for i in range(Parameters.ZER_points+1):
            l[i] = c[i]
            for j in range(10*Parameters.ZER_points+1):
                if  a[j] == 0.0:
                    continue
                bb = c[i] / a[j]
                if  bb < b_s.min_val or bb > b_s.max_val:
                    continue
                else:
                    mm = min(self.muValue(a[j]), b.muValue(bb)) # t-norm operator 
                    ml[i] = max(ml[i], mm)                      # s-norm operator 
        
        c = np.linspace(r_min, r_max, Parameters.ZER_points+1)
        mr = np.zeros(Parameters.ZER_points+1)     
        r =  np.zeros_like(mr)
        for i in range(Parameters.ZER_points+1):
            r[i] = c[i]
            for j in range(10*Parameters.ZER_points+1):
                if  a[j] == 0.0:
                    continue
                bb = c[i] / a[j]
                if  bb < b_s.min_val or bb > b_s.max_val:
                    continue
                else:
                    mm = min(self.muValue(a[j]), b.muValue(bb)) # t-norm operator 
                    mr[i] = max(mr[i], mm)                      # s-norm operator 
                                
        mr = mr[::-1]
        r = r[::-1]
        for i in range(Parameters.interp_points+1):
            f.left[i] = np.interp(f.mu[i],ml,l)
            f.right[i] = np.interp(f.mu[i],mr,r)
                       
        return(f)        
    
    
    # check if FN belongs to the interval [minv, maxv]
    def isInRange(self, minv, maxv):
        s = self.support()
        
        if  s.min_val >= minv and s.max_val <= maxv: return True
        else: return False 
        
 


##############################################################################################

def Singleton(point, name = ''):
    return FNumber(1, [point], name)

def Rectangle(points, name = ''):
    return FNumber(2, points, name)

def Triangle(points, name = ''):
    return FNumber(3, points, name)

def Trapezoid(points, name = ''):
    return FNumber(4, points, name)

def Polynomial(points, name = ''):
    return FNumber(5, points, name)

def Harmonic(points, name = ''):
    return FNumber(6, points, name)

def PseudoGaussian(points, name = ''):
    return FNumber(7, points, name)




##############################################################################################
# Similarity measure of 2 FNs
#
# sim_type = 1: sup_{X} mu_{A t_norm B}(x)
# sim_type = 2: int mu_{A t_norm B}(x) dx / int mu_{A s_norm B}(x) dx
# sim_type = 3: 1/(1+dist(f1,f2)
#
# t_norm: 'min', 'prod'
# s_norm: 'max', 'bsum'
##############################################################################################
def similarity_measure(f1, f2, sim_type=1, t_norm='min', s_norm='max'):

    if  isinstance(f1, numbers.Number):
        fn1 = Singleton(f1, '{0}'.format(f1))
    elif isinstance(f1, SInt.SInterval):
        fn1 = Rectangle([f1.min_val, f1.max_val], "[{0.min_val},{0.max_val}]".format(f1))
    elif isinstance(f1, FNumber):
        fn1  = f1
    else:
        raise NotImplemented()
    
    if  isinstance(f2, numbers.Number):
        fn2 = Singleton(f2, '{0}'.format(f2))
    elif isinstance(f2, SInt.SInterval):
        fn2 = Rectangle([f2.min_val, f2.max_val], "[{0.min_val},{0.max_val}]".format(f2))
    elif isinstance(f2, FNumber):
        fn2  = f2
    else:
        raise NotImplemented()
    
    if  sim_type == 1:
        w1 = SInt.intersection(fn1.core(), fn2.core())
        if  w1[0] == True:  return 1.0
        
        w2 = SInt.intersection(fn1.support(), fn2.support())
        if  w2[0] == False:  return 0.0
        
        if  fn1.fntype == 1:  return fn2.muValue(fn1.singleton_position) 
        if  fn2.fntype == 1:  return fn1.muValue(fn2.singleton_position)
        
        if  fn1.fntype == 2:  return max(fn2.muValue(fn1.left[0]), fn2.muValue(fn1.right[0]))
        if  fn2.fntype == 2:  return max(fn1.muValue(fn2.left[0]), fn1.muValue(fn2.right[0]))
        
        x_min = w2[1].min_val
        x_max = w2[1].max_val
        
        v_max = 0.0
        for x in np.linspace(x_min, x_max, Parameters.sim_interp_points+1):
            v1 = fn1.muValue(x)
            v2 = fn2.muValue(x)
            if  t_norm == 'min':   v = min(v1,v2)
            elif t_norm == 'prod': v = v1 * v2
            else: raise NotImplemented()
            
            if  v > v_max: v_max = v
        
        return v_max            
            
    elif sim_type == 2:
        w2 = SInt.intersection(fn1.support(), fn2.support())
        if  w2[0] == False:  return 0.0
        
        if  fn1.fntype == 1 or fn2.fntype == 1: return 0.0
        
        w1 = SInt.hull(fn1.support(), fn2.support())
        
        #numerator
        v_num = 0.0
        x_min = w2[1].min_val
        x_max = w2[1].max_val
        x = np.linspace(x_min, x_max, Parameters.sim_interp_points+1)
        dx = x[1] - x[0]
        for i in range(Parameters.sim_interp_points):
            v1 = fn1.muValue(x[i])
            v2 = fn2.muValue(x[i])
            if  t_norm == 'min':   v = min(v1,v2)
            elif t_norm == 'prod': v = v1 * v2
            else: raise NotImplemented()
            
            v1a = fn1.muValue(x[i+1])
            v2a = fn2.muValue(x[i+1])
            if  t_norm == 'min':   va = min(v1a,v2a)
            elif t_norm == 'prod': va = v1a * v2a
            else: raise NotImplemented()
            
            v_num += 0.5*(v + va)*dx
            
        #denominator
        v_den = 0.0
        x_min = w1.min_val
        x_max = w1.max_val
        x = np.linspace(x_min, x_max, Parameters.sim_interp_points+1)
        dx = x[1] - x[0]
        for i in range(Parameters.sim_interp_points):
            v1 = fn1.muValue(x[i])
            v2 = fn2.muValue(x[i])
            if  s_norm == 'max':   v = max(v1,v2)
            elif s_norm == 'bsum': v = min(1.0, v1 + v2)
            else: raise NotImplemented()
            
            v1a = fn1.muValue(x[i+1])
            v2a = fn2.muValue(x[i+1])
            if  s_norm == 'max':   va = max(v1a,v2a)
            elif s_norm == 'bsum': va = min(1.0, v1a + v2a)
            else: raise NotImplemented()
            
            v_den += 0.5*(v + va)*dx    
            
        if v_den == 0.0: return 0.0
        else: return v_num/v_den   
    
   
    elif sim_type == 3:
        return 1.0/(1.0 + distance(f1, f2, 3))
        
    else: raise NotImplemented()
                              



##############################################################################################
# Distance between 2 FNs
# dist_type = 1: distance between singleton positions
# dist_type = 2: distance between centers of gravity
# dist_type = 3: d = (\int (mu_a(x) - mu_b(x))**p dx)**(1.0/p)
# dist_type = 4: integral formula with p exponent 
# dist_type = 5: integral formula with max
##############################################################################################
def distance(f1, f2, dist_type=1, p=2):          

    if  isinstance(f1, numbers.Number):
        fn1 = Singleton(f1, '{0}'.format(f1))
    elif isinstance(f1, SInt.SInterval):
        fn1 = Rectangle([f1.min_val, f1.max_val], "[{0.min_val},{0.max_val}]".format(f1))
    elif isinstance(f1, FNumber):
        fn1  = f1
    else:
        raise NotImplemented()
    
    if  isinstance(f2, numbers.Number):
        fn2 = Singleton(f2, '{0}'.format(f2))
    elif isinstance(f2, SInt.SInterval):
        fn2 = Rectangle([f2.min_val, f2.max_val], "[{0.min_val},{0.max_val}]".format(f2))
    elif isinstance(f2, FNumber):
        fn2  = f2
    else:
        raise NotImplemented()
    
    if  dist_type == 1:        
        return abs(fn1.singleton_position - fn2.singleton_position)
        
    elif dist_type == 2:        
        return abs(fn1.center_of_gravity() - fn2.center_of_gravity())
    
    elif dist_type == 3:
        if  fn1.fntype == 1 and fn2.fntype == 1: return 0.0
        
        w1 = SInt.hull(fn1.support(), fn2.support())
        d = 0.0
        x_min = w1.min_val
        x_max = w1.max_val
        x = np.linspace(x_min, x_max, Parameters.sim_interp_points+1)
        dx = x[1] - x[0]
        for i in range(Parameters.sim_interp_points):
            v1 = fn1.muValue(x[i])
            v2 = fn2.muValue(x[i])
            v = np.fabs(v1 - v2)**p
            
            
            v1a = fn1.muValue(x[i+1])
            v2a = fn2.muValue(x[i+1])
            va = abs(v1a - v2a)**p
            
            d += 0.5*(v + va)*dx
   
        return d ** (1.0/p)       
    
    elif dist_type == 4 or dist_type == 5:
        d1 = 0.0
        d2 = 0.0
        
        x = np.linspace(0.0, 1.0, Parameters.sim_interp_points+1)
        dx = x[1] - x[0]
        for i in range(Parameters.sim_interp_points):
            
            a_fn1 = fn1.alphaValue(x[i])
            a_fn2 = fn2.alphaValue(x[i])
            v_left = abs(a_fn1.min_val - a_fn2.min_val)**p
            v_right = abs(a_fn1.max_val - a_fn2.max_val)**p
            
            a_fn1_1 = fn1.alphaValue(x[i+1])
            a_fn2_1 = fn2.alphaValue(x[i+1])
            v_left_1 = abs(a_fn1_1.min_val - a_fn2_1.min_val)**p
            v_right_1 = abs(a_fn1_1.max_val - a_fn2_1.max_val)**p
            
            d1 += 0.5*(v_left + v_left_1)*dx
            d2 += 0.5*(v_right + v_right_1)*dx
   
        if  dist_type == 4:
            return (0.5*d1 + 0.5*d2) ** (1.0/p)
        else: 
            return max(d1**(1.0/p), d2**(1.0/p))
    
    else:
        raise NotImplemented()




##############################################################################################
# Calculate width of mean symmetric triangular fuzzy number for list of different fuzzy numbers
##############################################################################################
def mean_width(y):
    sum = 0.0
    x = np.linspace(0.0,1.0,Parameters.sim_interp_points + 1)
    dx = x[1] - x[0]
    for k in range(len(y)):
        si = 0.0
        for i in range(Parameters.sim_interp_points):
            a_fn = y[k].alphaValue(x[i])
            a_fn_1 = y[k].alphaValue(x[i + 1])
            si += 0.5 * ((1.0-x[i+1])*(a_fn_1.max_val - a_fn_1.min_val) + (1.0-x[i])*(a_fn.max_val - a_fn.min_val)) * dx
        sum += si
    return 1.5 * sum/len(y)

