import numbers
import numpy as np

class SInterval:
    '''
    Simple Interval Class
    '''

    def __init__(self, min_val = 0, max_val = 1):
        if  min_val > max_val:
            raise ValueError("Lower bound of an interval can't be greater than upper!") 
        self.min_val = float(min_val)
        self.max_val = float(max_val)
    
        
    def __str__(self):
        return "[{0.min_val},{0.max_val}]".format(self)
    
    
    def __repr__(self):
        return self.__str__()         
    
    
    def __eq__(self, other):
        return self.min_val == other.min_val and self.max_val == other.max_val


    def __neg__(self):
        return SInterval(-self.max_val, -self.min_val)
    
    
    def __add__(self, other):
        if  isinstance(other, SInterval):
            return SInterval(self.min_val + other.min_val, self.max_val + other.max_val)
        if  isinstance(other,numbers.Number):
            return SInterval(self.min_val + other, self.max_val + other)
        raise NotImplemented()
    
    def __radd__(self, other):
        if  isinstance(other,numbers.Number):
            return SInterval(self.min_val + other, self.max_val + other)
        raise NotImplemented()      
                
                
    def __sub__(self, other):
        if  isinstance(other, SInterval):
            return SInterval(self.min_val - other.max_val, self.max_val - other.min_val)
        if  isinstance(other,numbers.Number):
            return SInterval(self.min_val - other, self.max_val - other)  
        raise NotImplemented()      
        
    def __rsub__(self, other):
        if  isinstance(other,numbers.Number):
            return SInterval(other - self.max_val, other - self.min_val)
        raise NotImplemented()    
    
    
    def __mul__(self, other):
        if  isinstance(other,numbers.Number):
            d = SInterval(other,other)
        elif isinstance(other, SInterval):
            d = other
        else:
            raise NotImplemented()
        sm = [self.min_val*d.min_val, self.min_val*d.max_val, self.max_val*d.min_val, self.max_val*d.max_val]
        return SInterval(min(sm), max(sm))
         
    
    def __rmul__(self, other):
        if  isinstance(other,numbers.Number):
            if  other > 0.0:
                return SInterval(self.min_val * other, self.max_val * other)  
            else:
                return SInterval(self.max_val * other, self.min_val * other)
        raise NotImplemented()
    
    
    def __truediv__(self, other):
        if  isinstance(other, SInterval):
            if  other.isSubset(0.0):
                raise ValueError("0 can't belong to an interval in denominator!") 
            d = SInterval(1.0/other.max_val, 1.0/other.min_val)
            sm = [self.min_val*d.min_val, self.min_val*d.max_val, self.max_val*d.min_val, self.max_val*d.max_val]
            return SInterval(min(sm), max(sm))
        if  isinstance(other,numbers.Number):
            if  other != 0.0:
                if  other > 0.0:
                    return SInterval(self.min_val / float(other), self.max_val / float(other))
                else:
                    return SInterval(self.max_val / float(other), self.min_val / float(other))
            else:
                raise ValueError('Division by 0!')  
        raise NotImplemented() 
    
    
    def __rtruediv__(self, other):
        if  isinstance(other,numbers.Number):
            if  self.isSubset(0.0):
                raise ValueError("0 can't belong to an interval in denominator!") 
            if  other > 0.0:
                return SInterval(other/self.max_val, other/self.min_val)
            else:  
                return SInterval(other/self.min_val, other/self.max_val)
        raise NotImplemented()
    
    
    def __abs__(self):
        if  self.min_val < 0.0 and self.max_val > 0.0:
            return SInterval(0.0, max([-self.min_val, self.max_val]))
        elif self.max_val <= 0.0:
            return SInterval(abs(self.max_val), abs(self.min_val))
        else:
            return self
        
        
    def __pow__(self, other):
        if  not isinstance(other,numbers.Number):
            raise ValueError("Exponent must be a number!") 
        
        if  other < 0 and self.isSubset(0.0):
            raise ValueError("Interval contains 0.0 and cannot be raised to a negative power")
        
        if  self.min_val >= 0.0: # positive interval
            if  other >= 0.0:
                return SInterval(self.min_val ** other, self.max_val ** other)
            else:
                return SInterval(self.max_val ** other, self.min_val ** other)
        else:
            if  int(other) != other:
                raise ValueError("Exponent must be an integer number if lower bound of an interval is negative!")
            else:
                l =  self.min_val ** other
                u =  self.max_val ** other
                if  self.max_val <= 0: # negative interval
                    return SInterval(min([l,u]), max([l,u]))
                else:
                    if  other % 2 == 0:
                        return SInterval(0.0, max([l,u]))
                    else:
                        return SInterval(min([l,u]), max([l,u]))
            
        
    
    # Test if d belongs to self interval
    def isSubset(self, d):
        if  isinstance(d, SInterval):
            if  d.min_val >= self.min_val and d.max_val <= self.max_val:
                return True
            else:
                return False
        if  isinstance(d,numbers.Number):
            if  d >= self.min_val and d <= self.max_val:
                return True
            else:
                return False 
        raise NotImplemented()
    
    
    def mid(self):
        return (self.min_val + self.max_val)/2.0
    
    
    def width(self):
        return float(self.max_val - self.min_val)

    
##############################################################################################    
    
    
def intersection(i1, i2):
    low = max(i1.min_val, i2.min_val)
    high = min(i1.max_val, i2.max_val)
    
    if  high >= low:
        return [True, SInterval(low, high)]
    else:
        return [False, SInterval(high, low)]
    


def hull(i1, i2):
    low = min(i1.min_val, i2.min_val)
    high = max(i1.max_val, i2.max_val)
    
    return SInterval(low, high)


def distance(i1, i2, dist_type=1):
    '''
    Distance between 2 intervals
    dist_type = 1: Hausdorff distance
    dist_type = 2: mean distance between lower and upper bounds
    dist_type = 3: distance between mid points
    '''
    if  isinstance(i1,numbers.Number):
        i1 = SInterval(i1,i1)
    if  isinstance(i2,numbers.Number):
        i2 = SInterval(i2,i2)    

    if  dist_type == 1:
        d = max([abs(i1.min_val-i2.min_val), abs(i1.max_val-i2.max_val)])
    elif dist_type == 2:
        d = 0.5*abs(i1.min_val-i2.min_val) + 0.5*abs(i1.max_val-i2.max_val)
    elif dist_type == 3:
        d = abs(i1.mid() - i2.mid())    
    return d


def midrad(x, rad=0.0):
    '''
    Transform numerical matrix into interval matrix.
    min_val = x - rad
    max_val = x + rad
    '''
    return infsup(x-rad, x+rad)


def infsup(x1, x2):
    '''
    Create interval matrix.
    x1 - lower bounds of intervals
    x2 - upper bounds of intervals
    '''
    m = []
    if  np.ndim(x1) == 1:
        w = np.size(x1)
        for i in range(w):
            m.append(SInterval(x1[i],x2[i]))
            
        return np.array(m, ndmin=1)    
    else:
        w = np.size(x1,0)
        k = np.size(x1,1)
        for i in range(w):
            r = []
            for j in range(k): 
                r.append(SInterval(x1[i,j],x2[i,j]))
            m.append(r)
        
        return np.array(m, ndmin=2)
    
    

